from typing import Literal

from astropy.io import fits
from astropy.time import Time
from astropy.time import TimeDelta
from dkist_processing_common.tasks import WriteL1Frame


class WriteL1Data(WriteL1Frame):
    def add_dataset_headers(
        self, header: fits.Header, stokes: Literal["I", "Q", "U", "V"]
    ) -> fits.Header:
        header["DAAXES"] = 2
        header["DEAXES"] = 1
        header["DNAXIS"] = 3
        header["LEVEL"] = 1
        header["WAVEMAX"] = 124
        header["WAVEMIN"] = 123
        header["WAVEREF"] = "Air"
        header["WAVEUNIT"] = -9
        header["DINDEX3"] = 3
        header["DNAXIS1"] = header["NAXIS1"]
        header["DNAXIS2"] = header["NAXIS2"]
        header["DNAXIS3"] = 10
        header["DPNAME1"] = "spatial x"
        header["DPNAME2"] = "spatial y"
        header["DPNAME3"] = "frame number"
        header["DTYPE1"] = "SPATIAL"
        header["DTYPE2"] = "SPATIAL"
        header["DTYPE3"] = "TEMPORAL"
        header["DUNIT1"] = "arcsec"
        header["DUNIT2"] = "arcsec"
        header["DUNIT3"] = "s"
        header["DWNAME1"] = "helioprojective longitude"
        header["DWNAME2"] = "helioprojective latitude"
        header["DWNAME3"] = "time"
        header["NBIN"] = 1
        for i in range(1, header["NAXIS"] + 1):
            header[f"NBIN{i}"] = 1

        return header

    def _calculate_date_end(self, header: fits.Header) -> str:
        """
        Calculate the VBI specific version of the "DATE-END" keyword.

        Parameters
        ----------
        header
            The input fits header

        Returns
        -------
        The isot formatted string of the DATE-END keyword value
        """
        return (
            Time(header["DATE-BEG"], format="isot", precision=6)
            + TimeDelta(float(header["TEXPOSUR"]) / 1000, format="sec")
        ).to_value("isot")



from datetime import datetime

import numpy as np
import pytest
from dkist_processing_common.tasks import WorkflowTaskBase
from dkist_processing_common.tasks.mixin.input_dataset import InputDatasetParameterValue

from dkist_processing_test.models.parameters import TestParameters


@pytest.fixture(scope="session")
def parameter_dict_with_path(tmp_path_factory, random_parameter_hdulist, parameter_file_object_key):
    """Enough of an input dataset parameters part to exercise file loading parameters."""
    hdul, mu, std, const = random_parameter_hdulist
    file_path = tmp_path_factory.mktemp("parameters") / parameter_file_object_key
    hdul.writeto(file_path)

    value = {
        "bucket": "raw",
        "objectKey": parameter_file_object_key,
        "param_path": file_path,
        "is_file": True,
    }

    param_dict = {
        "test_random_data": [
            InputDatasetParameterValue(
                parameter_value_id=1,
                parameter_value=value,
                parameter_value_start_date=datetime(1946, 11, 20),
            )
        ]
    }

    return param_dict, mu, std, const


@pytest.fixture(scope="session")
def task_with_parameters(parameter_dict_with_path):
    param_dict = parameter_dict_with_path[0]

    class TaskWithParameters(WorkflowTaskBase):
        def __init__(
            self,
            recipe_run_id: int,
            workflow_name: str,
            workflow_version: str,
        ):
            super().__init__(
                recipe_run_id=recipe_run_id,
                workflow_name=workflow_name,
                workflow_version=workflow_version,
            )
            self.parameters = TestParameters(param_dict)

        def run(self):
            """Do stuff."""
            pass

    task = TaskWithParameters(
        recipe_run_id=0,
        workflow_name="do_stuff",
        workflow_version="VX.Y",
    )

    return task


def test_parameter(task_with_parameters, parameter_dict_with_path):
    """
    Given: A task with parameters that depend on files
    When: Accessing those parameters
    Then: The correct values are returned
    """
    task = task_with_parameters
    _, mu, std, const = parameter_dict_with_path

    assert type(task.parameters.randomness) is tuple
    np.testing.assert_allclose(np.array(task.parameters.randomness), np.array([mu, std]), rtol=1)

    assert task.parameters.constant == const



"""
Workflow which exercises the api's but doesn't perform an action
"""
from dkist_processing_core import Workflow

from dkist_processing_test.tasks.noop import NoOpTask
from dkist_processing_test.tasks.noop import NoOpTask2


noop = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="noop",
    workflow_package=__package__,
)
noop.add_node(task=NoOpTask, upstreams=None)


noop_flow = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="noop-flow",
    workflow_package=__package__,
)
noop_flow.add_node(task=NoOpTask, upstreams=None)
noop_flow.add_node(task=NoOpTask2, upstreams=NoOpTask)



"""Parse Task definition."""
from typing import TypeVar

from dkist_processing_common.models.constants import BudName
from dkist_processing_common.models.flower_pot import SpilledDirt
from dkist_processing_common.models.flower_pot import Stem
from dkist_processing_common.models.flower_pot import Thorn
from dkist_processing_common.models.tags import StemName
from dkist_processing_common.parsers.l0_fits_access import L0FitsAccess
from dkist_processing_common.parsers.single_value_single_key_flower import (
    SingleValueSingleKeyFlower,
)
from dkist_processing_common.parsers.unique_bud import UniqueBud
from dkist_processing_common.tasks import ParseL0InputData

S = TypeVar("S", bound=Stem)


class TotalDspsRepeatsBud(UniqueBud):
    def __init__(self):
        super().__init__(
            constant_name=BudName.num_dsps_repeats.value, metadata_key="num_dsps_repeats"
        )

    def setter(self, fits_obj: L0FitsAccess):
        if fits_obj.ip_task_type != "observe":
            return SpilledDirt
        return super().setter(fits_obj)


class DspsRepeatNumberFlower(SingleValueSingleKeyFlower):
    def __init__(self):
        super().__init__(
            tag_stem_name=StemName.dsps_repeat.value, metadata_key="current_dsps_repeat"
        )

    def setter(self, fits_obj: L0FitsAccess):
        if fits_obj.ip_task_type != "observe":
            return SpilledDirt
        return super().setter(fits_obj)


class PickyDummyBud(Stem):
    """Exists to do literally nothing"""

    def setter(self, fits_obj: L0FitsAccess):
        if fits_obj.ip_task_type == "bad value":
            raise ValueError("This task type is bad!")

    def getter(self, key):
        return Thorn


class ParseL0TestInputData(ParseL0InputData):
    @property
    def fits_parsing_class(self):
        return L0FitsAccess

    @property
    def tag_flowers(self) -> list[S]:
        return super().tag_flowers + [
            SingleValueSingleKeyFlower(
                tag_stem_name=StemName.task.value, metadata_key="ip_task_type"
            ),
            DspsRepeatNumberFlower(),
        ]

    @property
    def constant_flowers(self) -> list[S]:
        return super().constant_flowers + [
            TotalDspsRepeatsBud(),
            PickyDummyBud(stem_name="PICKY_BUD"),
        ]



"""
Workflow which exercises the common tasks in an end to end scenario
"""
from dkist_processing_common.tasks import AddDatasetReceiptAccount
from dkist_processing_common.tasks import PublishCatalogAndQualityMessages
from dkist_processing_common.tasks import QualityL1Metrics
from dkist_processing_common.tasks import Teardown
from dkist_processing_common.tasks import TransferL0Data
from dkist_processing_common.tasks import TransferL1Data
from dkist_processing_core import Workflow

from dkist_processing_test.tasks.fake_science import GenerateCalibratedData
from dkist_processing_test.tasks.movie import AssembleTestMovie
from dkist_processing_test.tasks.movie import MakeTestMovieFrames
from dkist_processing_test.tasks.parse import ParseL0TestInputData
from dkist_processing_test.tasks.quality import TestQualityL0Metrics
from dkist_processing_test.tasks.quality import TestSubmitQuality
from dkist_processing_test.tasks.write_l1 import WriteL1Data

# TransferInputData Task
transfer_input_data = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="transfer-input-data",
    workflow_package=__package__,
)
transfer_input_data.add_node(task=TransferL0Data, upstreams=None)

# ParseInputData Task
parse_input_data = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="parse-input-data",
    workflow_package=__package__,
)
parse_input_data.add_node(task=ParseL0TestInputData, upstreams=None)

# L0Quality Task
quality_l0_metrics = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="quality-l0-metrics",
    workflow_package=__package__,
)
quality_l0_metrics.add_node(task=TestQualityL0Metrics, upstreams=None)

# L1Quality Task
quality_l1_metrics = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="quality-l1-metrics",
    workflow_package=__package__,
)
quality_l1_metrics.add_node(task=QualityL1Metrics, upstreams=None)

# SubmitQuality Task
quality_submit_metrics = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="quality-submit-metrics",
    workflow_package=__package__,
)
quality_submit_metrics.add_node(task=TestSubmitQuality, upstreams=None)

# GenerateL1CalibratedData Task
generate_calibrated_data = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="generate-calibrated-data",
    workflow_package=__package__,
)
generate_calibrated_data.add_node(task=GenerateCalibratedData, upstreams=None)

# MakeTestMovieFrames task
make_test_movie_frames = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="make-test-movie-frames",
    workflow_package=__package__,
)
make_test_movie_frames.add_node(task=MakeTestMovieFrames, upstreams=None)

# AssembleTestMovie Task
assemble_test_movie = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="assemble-test-movie",
    workflow_package=__package__,
)
assemble_test_movie.add_node(task=AssembleTestMovie, upstreams=None)

# WriteL1 Task
write_l1 = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="write-l1",
    workflow_package=__package__,
)
write_l1.add_node(task=WriteL1Data, upstreams=None)

# TransferOutputData Task
transfer_output_data = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="transfer-output-data",
    workflow_package=__package__,
)
transfer_output_data.add_node(task=TransferL1Data, upstreams=None)

# AddDatasetReceiptAccount Task
add_dataset_receipt_account = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="add-dataset-receipt-account",
    workflow_package=__package__,
)
add_dataset_receipt_account.add_node(task=AddDatasetReceiptAccount, upstreams=None)

# PublishCatalogMessages Task
publish_catalog_messages = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="publish-messages",
    workflow_package=__package__,
)
publish_catalog_messages.add_node(task=PublishCatalogAndQualityMessages, upstreams=None)

# Teardown Task
teardown = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="teardown",
    workflow_package=__package__,
)
teardown.add_node(task=Teardown, upstreams=None)



"""
Workflow which exercises the common tasks in an end to end scenario
"""
from dkist_processing_common.tasks import AddDatasetReceiptAccount
from dkist_processing_common.tasks import PublishCatalogAndQualityMessages
from dkist_processing_common.tasks import QualityL1Metrics
from dkist_processing_common.tasks import SubmitQuality
from dkist_processing_common.tasks import Teardown
from dkist_processing_common.tasks import TransferL0Data
from dkist_processing_common.tasks import TransferL1Data
from dkist_processing_core import Workflow

from dkist_processing_test.tasks.fake_science import GenerateCalibratedData
from dkist_processing_test.tasks.movie import AssembleTestMovie
from dkist_processing_test.tasks.movie import MakeTestMovieFrames
from dkist_processing_test.tasks.parse import ParseL0TestInputData
from dkist_processing_test.tasks.quality import TestQualityL0Metrics
from dkist_processing_test.tasks.write_l1 import WriteL1Data

end_to_end = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="management-processes-e2e",
    workflow_package=__package__,
)
end_to_end.add_node(task=TransferL0Data, upstreams=None)
end_to_end.add_node(task=ParseL0TestInputData, upstreams=TransferL0Data)
end_to_end.add_node(task=TestQualityL0Metrics, upstreams=ParseL0TestInputData)
end_to_end.add_node(task=GenerateCalibratedData, upstreams=TestQualityL0Metrics)
end_to_end.add_node(task=WriteL1Data, upstreams=GenerateCalibratedData)
end_to_end.add_node(task=QualityL1Metrics, upstreams=WriteL1Data)
end_to_end.add_node(task=SubmitQuality, upstreams=QualityL1Metrics)
end_to_end.add_node(task=MakeTestMovieFrames, upstreams=WriteL1Data)
end_to_end.add_node(task=AssembleTestMovie, upstreams=MakeTestMovieFrames)
end_to_end.add_node(task=AddDatasetReceiptAccount, upstreams=[AssembleTestMovie, SubmitQuality])
end_to_end.add_node(task=TransferL1Data, upstreams=AddDatasetReceiptAccount)
end_to_end.add_node(task=PublishCatalogAndQualityMessages, upstreams=TransferL1Data)
end_to_end.add_node(task=Teardown, upstreams=PublishCatalogAndQualityMessages)



"""
Fake science task
"""
import numpy as np
from astropy.io import fits
from dkist_processing_common.models.tags import Tag
from dkist_processing_common.tasks import WorkflowTaskBase
from dkist_processing_common.tasks.mixin.fits import FitsDataMixin
from dkist_processing_common.tasks.mixin.input_dataset import InputDatasetMixin

from dkist_processing_test.models.parameters import TestParameters


class GenerateCalibratedData(WorkflowTaskBase, FitsDataMixin, InputDatasetMixin):

    record_provenance = True

    def __init__(
        self,
        recipe_run_id: int,
        workflow_name: str,
        workflow_version: str,
    ):
        super().__init__(
            recipe_run_id=recipe_run_id,
            workflow_name=workflow_name,
            workflow_version=workflow_version,
        )
        self.parameters = TestParameters(self.input_dataset_parameters)

    def run(self):
        rng = np.random.default_rng()
        count = 1  # keep a running count to increment the dsps repeat number
        with self.apm_task_step("Looping over inputs"):
            for path, hdu in self.fits_data_read_hdu(tags=Tag.input()):
                header = hdu.header
                with self.apm_processing_step("Doing some calculations"):
                    header["DSPSNUM"] = count
                    data = hdu.data

                    # Just do some weird crap. We don't use the loaded random array directly so that we
                    # don't have to care that the shapes are the same as the "real" data.
                    random_signal = rng.normal(*self.parameters.randomness, size=data.shape)
                    data = (
                        data + random_signal
                    )  # Needs to be like this because data will start as int-type
                    data += self.parameters.constant
                    output_hdu = fits.PrimaryHDU(data=data, header=header)

                with self.apm_writing_step("Writing data"):
                    output_hdul = fits.HDUList([output_hdu])
                    self.fits_data_write(
                        hdu_list=output_hdul,
                        tags=[
                            Tag.calibrated(),
                            Tag.frame(),
                            Tag.stokes("I"),
                            Tag.dsps_repeat(count),
                        ],
                    )
                count += 1



import numpy as np
from astropy.io import fits
from dkist_processing_common.models.parameters import ParameterBase
from dkist_processing_common.tasks.mixin.input_dataset import InputDatasetParameterValue


class TestParameters(ParameterBase):
    """Class to test loading parameters from a file."""

    def __init__(
        self,
        input_dataset_parameters: dict[str, list[InputDatasetParameterValue]],
    ):
        super().__init__(input_dataset_parameters)

    @property
    def randomness(self) -> (float, float):
        """A dummy parameter that requires loading a file."""
        with open(self._find_most_recent_past_value("test_random_data")["param_path"], "rb") as f:
            data = fits.open(f)[0].data
            mean = np.nanmean(data)
            std = np.nanstd(data)

        return mean, std

    @property
    def constant(self) -> float:
        """A dummy parameter that depends on the same file as a different parameter."""
        with open(self._find_most_recent_past_value("test_random_data")["param_path"], "rb") as f:
            data = fits.open(f)[1].data
            constant = np.median(data)

        return float(constant)



"""
Test task for infrastructure integration that will always fail
"""
from dkist_processing_core import TaskBase


class FailTask(TaskBase):
    def run(self) -> None:
        raise RuntimeError("Failure is guaranteed")



"""
Fake MakeMovieFrames and AssembleTestMovie
"""
from astropy.io import fits
from dkist_processing_common.models.tags import Tag
from dkist_processing_common.parsers.l1_fits_access import L1FitsAccess
from dkist_processing_common.tasks import AssembleMovie
from dkist_processing_common.tasks import WorkflowTaskBase
from dkist_processing_common.tasks.mixin.fits import FitsDataMixin
from PIL import ImageDraw


class MakeTestMovieFrames(WorkflowTaskBase, FitsDataMixin):
    """
    Take each output frame, copy the header and data and write out
    as a movie frame
    """

    def run(self):
        for d in range(1, self.constants.num_dsps_repeats + 1):
            with self.apm_task_step(f"Workign on dsps repeat {d}"):
                for path, hdu in self.fits_data_read_hdu(tags=[Tag.output(), Tag.dsps_repeat(d)]):
                    header = hdu.header
                    data = hdu.data
                    output_hdu = fits.PrimaryHDU(data=data, header=header)
                    output_hdul = fits.HDUList([output_hdu])

                    with self.apm_writing_step("Writing data"):
                        self.fits_data_write(
                            hdu_list=output_hdul,
                            tags=[Tag.movie_frame(), Tag.dsps_repeat(d)],
                        )


class AssembleTestMovie(AssembleMovie):
    """
    A shell to extend the AssembleMovie class for the end-to-end test.
    """

    @property
    def fits_parsing_class(self):
        return L1FitsAccess

    def write_overlay(self, draw: ImageDraw, fits_obj: L1FitsAccess) -> None:
        pass



"""
Workflow which is designed to fail
"""
from dkist_processing_core import Workflow

from dkist_processing_test.tasks.fail import FailTask


fail = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="fail",
    workflow_package=__package__,
)
fail.add_node(task=FailTask, upstreams=None)



"""Global test fixture configuration"""
from random import randint

import numpy as np
import pytest
from astropy.io import fits
from dkist_data_simulator.spec122 import Spec122Dataset
from dkist_header_validator.translator import translate_spec122_to_spec214_l0


@pytest.fixture()
def recipe_run_id():
    return randint(0, 99999)


class S122Headers(Spec122Dataset):
    def __init__(
        self,
        array_shape: tuple[int, ...],
        num_steps: int = 4,
        num_exp_per_step: int = 1,
        num_dsps_repeats: int = 5,
        time_delta: float = 10.0,
        instrument: str = "vbi",
    ):
        dataset_shape = (num_exp_per_step * num_steps * num_dsps_repeats,) + array_shape[-2:]
        super().__init__(
            dataset_shape=dataset_shape,
            array_shape=array_shape,
            time_delta=time_delta,
            instrument=instrument,
        )
        self.num_steps = num_steps
        self.num_exp_per_step = num_exp_per_step
        self.num_dsps_repeats = num_dsps_repeats


def generate_214_l0_fits_frame(
    s122_header: fits.Header, data: np.ndarray | None = None
) -> fits.HDUList:
    """Convert S122 header into 214 L0"""
    if data is None:
        data = np.ones((1, 10, 10))
    translated_header = translate_spec122_to_spec214_l0(s122_header)
    del translated_header["COMMENT"]
    hdu = fits.PrimaryHDU(data=data, header=fits.Header(translated_header))
    return fits.HDUList([hdu])


@pytest.fixture(scope="session")
def parameter_file_object_key() -> str:
    return "random.fits"


@pytest.fixture(scope="session")
def random_parameter_hdulist() -> (fits.HDUList, float, float, float):
    rng = np.random.default_rng()
    mu, std = 10.0, 2.0
    const = 5.0
    rand_data = rng.normal(mu, std, size=(100, 100))
    const_data = np.ones((10, 10)) * const
    hdul = fits.HDUList([fits.PrimaryHDU(rand_data), fits.ImageHDU(const_data)])

    return hdul, mu, std, const



"""
Basic exercising of numba
"""
import logging
import timeit

import numpy as np
from dkist_processing_common.tasks import WorkflowTaskBase
from numba import njit


class ExerciseNumba(WorkflowTaskBase):
    def run(self):
        bubblesort_numba = njit(self.bubblesort)
        original = np.linspace(0.0, 10.0, 1001)
        shuffled_1 = original.copy()
        np.random.shuffle(shuffled_1)
        shuffled_2 = shuffled_1.copy()
        foo_1 = timeit.Timer(lambda: self.bubblesort(shuffled_1), globals=globals())
        time_1 = foo_1.timeit(100)
        foo_2 = timeit.Timer(lambda: bubblesort_numba(shuffled_2), globals=globals())
        time_2 = foo_2.timeit(100)
        speedup = time_1 / time_2
        logging.info(f"Normal task execution time: {time_1} secs")
        logging.info(f"Numba task execution time: {time_2} secs")
        logging.info(f"ExerciseNumba: Achieved a speedup of {speedup} using numba.")
        self.speedup = speedup
        self.sorted_array = shuffled_2

    @staticmethod
    def bubblesort(x):
        """Simple bubblesort algorithm copied from numba documentation"""
        n = len(x)
        for end in range(n, 1, -1):
            for i in range(end - 1):
                cur = x[i]
                if cur > x[i + 1]:
                    tmp = x[i]
                    x[i] = x[i + 1]
                    x[i + 1] = tmp



"""Test integrity of workflows."""
from dkist_processing_core.build_utils import validate_workflows

from dkist_processing_test import workflows


def test_workflow_integrity():
    """Validate workflow to ensure acyclic-ness and export compilation"""
    validate_workflows(workflows)



# TransferInputData Task
from dkist_processing_core import Workflow

from dkist_processing_test.tasks.exercise_numba import ExerciseNumba

exercise_numba = Workflow(
    input_data="input",
    output_data="output",
    category="test",
    detail="exercise_numba",
    workflow_package=__package__,
)
exercise_numba.add_node(task=ExerciseNumba, upstreams=None)



"""Quality task definition."""
from dkist_processing_common.models.tags import Tag
from dkist_processing_common.tasks import QualityL0Metrics
from dkist_processing_common.tasks import SubmitQuality


class TestQualityL0Metrics(QualityL0Metrics):
    def run(self) -> None:
        paths = self.read(tags=[Tag.input()])
        self.calculate_l0_metrics(paths=paths)


class TestSubmitQuality(SubmitQuality):
    @property
    def polcal_label_list(self) -> list[str] | None:
        return ["Beam1"]



"""
Tests for the tasks defined in this repo
"""
import json
import logging
from dataclasses import asdict
from dataclasses import dataclass
from datetime import datetime
from random import randint

import numpy as np
import pytest
from astropy.io import fits
from dkist_data_simulator.spec122 import Spec122Dataset
from dkist_header_validator import spec122_validator
from dkist_processing_common._util.scratch import WorkflowFileSystem
from dkist_processing_common.models.tags import Tag
from dkist_processing_common.tests.conftest import FakeGQLClient

from dkist_processing_test.models.parameters import TestParameters
from dkist_processing_test.tasks.exercise_numba import ExerciseNumba
from dkist_processing_test.tasks.fail import FailTask
from dkist_processing_test.tasks.fake_science import GenerateCalibratedData
from dkist_processing_test.tasks.movie import AssembleTestMovie
from dkist_processing_test.tasks.movie import MakeTestMovieFrames
from dkist_processing_test.tasks.noop import NoOpTask
from dkist_processing_test.tasks.write_l1 import WriteL1Data
from dkist_processing_test.tests.conftest import generate_214_l0_fits_frame
from dkist_processing_test.tests.conftest import S122Headers


@dataclass
class FakeConstantDb:
    NUM_DSPS_REPEATS: int = 2
    INSTRUMENT: str = "TEST"
    AVERAGE_CADENCE: float = 10.0
    MINIMUM_CADENCE: float = 10.0
    MAXIMUM_CADENCE: float = 10.0
    VARIANCE_CADENCE: float = 0.0
    STOKES_PARAMS: tuple[str] = (
        "I",
        "Q",
        "U",
        "V",
    )  # A tuple because lists aren't allowed on dataclasses


@pytest.fixture()
def noop_task():
    return NoOpTask(recipe_run_id=1, workflow_name="noop", workflow_version="VX.Y")


def test_noop_task(noop_task):
    """
    Given: A NoOpTask
    When: Calling the task instance
    Then: No errors raised
    """
    noop_task()


@pytest.fixture()
def fail_task():
    return FailTask(recipe_run_id=1, workflow_name="fail", workflow_version="VX.Y")


def test_fail_task(fail_task):
    """
    Given: A FailTask
    When: Calling the task instance
    Then: Runtime Error raised
    """
    with pytest.raises(RuntimeError):
        fail_task()


@pytest.fixture()
def generate_calibrated_data_task(
    tmp_path,
    recipe_run_id,
    assign_input_dataset_doc_to_task,
    parameter_file_object_key,
    random_parameter_hdulist,
):
    number_of_frames = 10
    with GenerateCalibratedData(
        recipe_run_id=recipe_run_id, workflow_name="GenerateCalibratedData", workflow_version="VX.Y"
    ) as task:
        # configure input data
        task.scratch = WorkflowFileSystem(scratch_base_path=tmp_path, recipe_run_id=recipe_run_id)
        input_frame_set = Spec122Dataset(
            instrument="vbi",
            dataset_shape=(number_of_frames, 512, 512),
            array_shape=(1, 512, 512),
            time_delta=10,
        )
        # load input data
        for idx, input_frame in enumerate(input_frame_set):
            hdu = input_frame.hdu()
            hdu.data = (
                np.ones(hdu.data.shape, dtype=int) * 10
            )  # Because input data will be ints in test system
            hdu.header["DSPSNUM"] = 1
            hdul = fits.HDUList([hdu])
            file_name = f"input_{idx}.fits"
            task.fits_data_write(hdu_list=hdul, tags=Tag.input(), relative_path=file_name)

        # Write parameter file
        hdul = random_parameter_hdulist[0]
        task.fits_data_write(hdu_list=hdul, tags=Tag.parameter(parameter_file_object_key))

        # This needs to be after we've tagged the parameter file
        assign_input_dataset_doc_to_task(task)

        # result
        yield task, number_of_frames
        # teardown
        task.scratch.purge()
        task.constants._purge()
    # disconnect


@pytest.fixture(scope="session")
def input_dataset_document_parameters_part_with_file(parameter_file_object_key):
    param_name = "test_random_data"
    value = {
        "__file__": {"bucket": "foo", "objectKey": parameter_file_object_key},
    }
    value_id = randint(1000, 2000)
    values = [
        {
            "parameterValueId": value_id,
            "parameterValue": json.dumps(value),
            "parameterValueStartDate": "1946-11-20",
        }
    ]
    parameter = {"parameterName": param_name, "parameterValues": values}
    parameters_list = [parameter]

    return parameters_list


@pytest.fixture(scope="session")
def assign_input_dataset_doc_to_task(input_dataset_document_parameters_part_with_file):
    def update_task(task):
        doc_path = task.scratch.workflow_base_path / "dataset_parameters.json"
        with open(doc_path, "w") as f:
            f.write(json.dumps(input_dataset_document_parameters_part_with_file))
        task.tag(doc_path, Tag.input_dataset_parameters())
        task.parameters = TestParameters(task.input_dataset_parameters)

    return update_task


def test_generate_calibrated_data(generate_calibrated_data_task, mocker):
    """
    Given: A GenerateCalibratedData task
    When: Calling the task instance
    Then: Output files are generated for each input file with appropriate tags
    """
    mocker.patch(
        "dkist_processing_common.tasks.mixin.metadata_store.GraphQLClient", new=FakeGQLClient
    )
    task, number_of_frames = generate_calibrated_data_task
    task()
    # Then
    calibrated_frame_paths = list(task.read(tags=[Tag.calibrated(), Tag.frame()]))

    # Verify frames
    assert len(calibrated_frame_paths) == number_of_frames
    for filepath in calibrated_frame_paths:
        assert filepath.exists()


class CommonDataset(Spec122Dataset):
    def __init__(self):
        super().__init__(
            array_shape=(1, 10, 10),
            time_delta=1,
            dataset_shape=(2, 10, 10),
            instrument="visp",
            start_time=datetime(2020, 1, 1, 0, 0, 0),
        )

        self.add_constant_key("TELEVATN", 6.28)
        self.add_constant_key("TAZIMUTH", 3.14)
        self.add_constant_key("TTBLANGL", 1.23)
        self.add_constant_key("INST_FOO", "bar")
        self.add_constant_key("DKIST004", "observe")
        self.add_constant_key("ID___005", "ip id")
        self.add_constant_key("PAC__004", "Sapphire Polarizer")
        self.add_constant_key("PAC__005", "31.2")
        self.add_constant_key("PAC__006", "clear")
        self.add_constant_key("PAC__007", "6.66")
        self.add_constant_key("PAC__008", "DarkShutter")
        self.add_constant_key("INSTRUME", "VISP")
        self.add_constant_key("WAVELNTH", 1080.0)
        self.add_constant_key("DATE-OBS", "2020-01-02T00:00:00.000")
        self.add_constant_key("DATE-END", "2020-01-03T00:00:00.000")
        self.add_constant_key("ID___013", "PROPOSAL_ID1")
        self.add_constant_key("PAC__002", "clear")
        self.add_constant_key("PAC__003", "on")
        self.add_constant_key("TELSCAN", "Raster")
        self.add_constant_key("DKIST008", 1)
        self.add_constant_key("DKIST009", 1)
        self.add_constant_key("BZERO", 0)
        self.add_constant_key("BSCALE", 1)


@pytest.fixture()
def complete_common_header():
    """
    A header with some common by-frame keywords
    """
    # Taken from dkist-processing-common
    ds = CommonDataset()
    header_list = [
        spec122_validator.validate_and_translate_to_214_l0(d.header(), return_type=fits.HDUList)[
            0
        ].header
        for d in ds
    ]

    return header_list[0]


@pytest.fixture(scope="function", params=[1, 4])
def write_l1_task(complete_common_header, request):
    with WriteL1Data(
        recipe_run_id=randint(0, 99999),
        workflow_name="workflow_name",
        workflow_version="workflow_version",
    ) as task:
        num_of_stokes_params = request.param
        stokes_params = ["I", "Q", "U", "V"]
        hdu = fits.PrimaryHDU(
            data=np.random.random(size=(1, 128, 128)) * 10, header=complete_common_header
        )
        logging.info(f"{num_of_stokes_params=}")
        hdul = fits.HDUList([hdu])
        for i in range(num_of_stokes_params):
            task.fits_data_write(
                hdu_list=hdul,
                tags=[Tag.calibrated(), Tag.frame(), Tag.stokes(stokes_params[i])],
            )
        task.constants._update(
            asdict(
                FakeConstantDb(
                    AVERAGE_CADENCE=10,
                    MINIMUM_CADENCE=10,
                    MAXIMUM_CADENCE=10,
                    VARIANCE_CADENCE=0,
                    INSTRUMENT="TEST",
                )
            )
        )
        yield task, num_of_stokes_params
        task.constants._purge()
        task.scratch.purge()


def test_write_l1_task(write_l1_task, mocker):
    """
    :Given: a write L1 task
    :When: running the task
    :Then: no errors are raised
    """
    mocker.patch(
        "dkist_processing_common.tasks.mixin.metadata_store.GraphQLClient", new=FakeGQLClient
    )
    task, num_of_stokes_params = write_l1_task
    task()
    files = list(task.read(tags=[Tag.frame(), Tag.output()]))
    logging.info(f"{files=}")
    assert len(files) == num_of_stokes_params
    for file in files:
        logging.info(f"Checking file {file}")
        assert file.exists


@pytest.fixture()
def make_movie_frames_task(tmp_path, recipe_run_id):
    with MakeTestMovieFrames(
        recipe_run_id=recipe_run_id, workflow_name="MakeMovieFrames", workflow_version="VX.Y"
    ) as task:
        task.scratch = WorkflowFileSystem(scratch_base_path=tmp_path, recipe_run_id=recipe_run_id)
        task.testing_num_dsps_repeats = 10
        task.num_steps = 1
        task.num_exp_per_step = 1
        task.constants._update(
            asdict(FakeConstantDb(NUM_DSPS_REPEATS=task.testing_num_dsps_repeats))
        )
        ds = S122Headers(
            array_shape=(1, 10, 10),
            num_steps=task.num_steps,
            num_exp_per_step=task.num_exp_per_step,
            num_dsps_repeats=task.testing_num_dsps_repeats,
        )
        header_generator = (d.header() for d in ds)
        for d, header in enumerate(header_generator):
            data = np.ones((10, 10))
            data[: d * 10, :] = 0.0
            hdl = generate_214_l0_fits_frame(data=data, s122_header=header)
            task.fits_data_write(
                hdu_list=hdl,
                tags=[
                    Tag.output(),
                    Tag.dsps_repeat(d + 1),
                ],
            )
        yield task
        task.scratch.purge()
        task.constants._purge()


def test_make_movie_frames_task(make_movie_frames_task, mocker):
    """
    :Given: a make_movie_frames_task task
    :When: running the task
    :Then: no errors are raised and a movie file is created
    """
    mocker.patch(
        "dkist_processing_common.tasks.mixin.metadata_store.GraphQLClient", new=FakeGQLClient
    )
    task = make_movie_frames_task
    task()
    movie_frames = list(task.read(tags=[Tag.movie_frame()]))
    logging.info(f"{movie_frames=}")
    assert len(movie_frames) == task.testing_num_dsps_repeats
    for frame in movie_frames:
        assert frame.exists()


@pytest.fixture()
def assemble_test_movie_task(tmp_path, recipe_run_id):
    with AssembleTestMovie(
        recipe_run_id=recipe_run_id, workflow_name="AssembleTestMovie", workflow_version="VX.Y"
    ) as task:
        task.scratch = WorkflowFileSystem(scratch_base_path=tmp_path)
        task.testing_num_dsps_repeats = 10
        task.num_steps = 1
        task.num_exp_per_step = 1
        task.constants._update(
            asdict(FakeConstantDb(NUM_DSPS_REPEATS=task.testing_num_dsps_repeats))
        )
        ds = S122Headers(
            array_shape=(1, 10, 10),
            num_steps=task.num_steps,
            num_exp_per_step=task.num_exp_per_step,
            num_dsps_repeats=task.testing_num_dsps_repeats,
        )
        header_generator = (d.header() for d in ds)
        for d, header in enumerate(header_generator):
            data = np.ones((10, 10))
            data[: d * 10, :] = 0.0
            hdl = generate_214_l0_fits_frame(data=data, s122_header=header)
            task.fits_data_write(
                hdu_list=hdl,
                tags=[
                    Tag.movie_frame(),
                    Tag.dsps_repeat(d + 1),
                ],
            )
        yield task
        task.scratch.purge()
        task.constants._purge()


def test_assemble_test_movie_task(assemble_test_movie_task, mocker):
    """
    :Given: an assemble_test_movie task
    :When: running the task
    :Then: no errors are raised and a movie file is created
    """
    mocker.patch(
        "dkist_processing_common.tasks.mixin.metadata_store.GraphQLClient", new=FakeGQLClient
    )
    task = assemble_test_movie_task
    task()
    movie_file = list(task.read(tags=[Tag.movie()]))
    logging.info(f"{movie_file=}")
    assert len(movie_file) == 1
    assert movie_file[0].exists()


@pytest.fixture()
def exercise_numba_task(recipe_run_id):
    with ExerciseNumba(
        recipe_run_id=recipe_run_id, workflow_name="ExerciseNumba", workflow_version="VX.Y"
    ) as task:
        yield task


def test_exercise_numba_task(exercise_numba_task):
    """
    :Given: an exercise_numba task
    :When: running the task
    :Then: the numba module can be loaded and simple method using numba is executed
    """
    original = np.linspace(0.0, 10.0, 1001)
    task = exercise_numba_task
    task()
    assert task.speedup > 1.0
    assert np.all(np.equal(original, task.sorted_array))
